#include <cs50.h>
#include <stdio.h>
#include <string.h>

// Max number of candidates
#define MAX 9

// preferences[i][j] is number of voters who prefer i over j
int preferences[MAX][MAX];

// locked[i][j] means i is locked in over j
bool locked[MAX][MAX];

// Each pair has a winner, loser
typedef struct
{
    int winner;
    int loser;
}
pair;

// Array of candidates
string candidates[MAX];
pair pairs[MAX * (MAX - 1) / 2];

int pair_count;
int candidate_count;

// Function prototypes
bool vote(int rank, string name, int ranks[]);
void record_preferences(int ranks[]);
void add_pairs(void);
void sort_pairs(void);
void lock_pairs(void);
void print_winner(void);
bool cycle(int winner, int loser);


void swap(int *xp, int *yp)
{
    int temp = *xp;
    *xp = *yp;
    *yp = temp;
}

int main(int argc, string argv[])
{
    // Check for invalid usage
    if (argc < 2)
    {
        printf("Usage: tideman [candidate ...]\n");
        return 1;
    }

    // Populate array of candidates
    candidate_count = argc - 1;
    // printf("Candidate count: %i\n", candidate_count);
    if (candidate_count > MAX)
    {
        return 2;
    }
    for (int i = 0; i < candidate_count; i++)
    {
        candidates[i] = argv[i + 1];
    }

    // Clear graph of locked in pairs
    for (int i = 0; i < candidate_count; i++)
    {
        for (int j = 0; j < candidate_count; j++)
        {
            locked[i][j] = false;
        }
    }

    pair_count = 0;
    int voter_count = get_int("Number of voters: ");

    // Query for votes
    for (int i = 0; i < voter_count; i++)
    {
        // ranks[i] is voter's ith preference
        int ranks[candidate_count];

        // Query for each rank
        for (int j = 0; j < candidate_count; j++)
        {
            string name = get_string("Rank %i: ", j + 1);

            if (!vote(j, name, ranks))
            {
                printf("Invalid vote.\n");
                return 3;
            }
        }

        record_preferences(ranks);

        printf("\n");
    }

    add_pairs();
    sort_pairs();
    lock_pairs();
    print_winner();
    return 0;
}

// Update ranks given a new vote
bool vote(int rank, string name, int ranks[])
{
    // Loop through candidate names to see if there is a match
    for (int i = 0, j = candidate_count; i < j; i++)
    {
        if (strcmp(name, candidates[i]) == 0)
        {
            // Update the rank to specify that the given name is assigned to the current rank
            ranks[rank] = i;

            return true;
        }
    }
    return false;
}

// Update preferences given one voter's ranks
void record_preferences(int ranks[])
{
    for (int i = 0; i < candidate_count; i++)
    {
        for (int j = 0; j < candidate_count; j++)
        {
            // For all index places that are smaller than what they are being compared to, they are higher in rank
            if (i < j)
            {
                preferences[ranks[i]][ranks[j]]++;
            }
            if (i >= j)
            {
                // preferences[ranks[i]][ranks[j]];
            }
        }
    }
    return;
}

// Record pairs of candidates where one is preferred over the other
void add_pairs(void)
{
    for (int i = 0; i < candidate_count; i++)
    {
        for (int j = 0; j < candidate_count; j++)
        {
            if (preferences[i][j] != preferences[j][i])
            {
                // If i wins over j more often than j wins over i, then make i winner and j loser
                if (preferences[i][j] > preferences[j][i] && i != j)
                {
                    pair_count++;

                    pairs[pair_count - 1].winner = i;
                    pairs[pair_count - 1].loser = j;
                }
            }
        }
    }
    return;
}

// Sort pairs in decreasing order by strength of victory
void sort_pairs(void)
{
    pair swap;

    // Logic of selection sort algorithm
    for (int c = 0; c < pair_count; c++)
    {
        for (int d = c + 1; d < pair_count; d++)
        {
            if ((preferences[pairs[c].winner][pairs[c].loser] - preferences[pairs[c].loser][pairs[c].winner]) <
                (preferences[pairs[d].winner][pairs[d].loser] - preferences[pairs[d].loser][pairs[d].winner]))
            {
                swap = pairs[c];
                pairs[c] = pairs[d];
                pairs[d] = swap;
            }
        }
    }
    return;
}

// Lock pairs into the candidate graph in order, without creating cycles
void lock_pairs(void)
{
    // Loop through each sorted pair
    for (int i = 0; i < pair_count; i++)
    {
        // Call on recursive function to check if the current winner is actually also a loser - which would close the cycle
        // Since all pairs are set to false for locked we only need to change that to true if we find there is no cycle
        if (!cycle(pairs[i].winner, pairs[i].loser))
        {
            // If there's no cycle then set current pair to locked true - that's all!
            locked[pairs[i].winner][pairs[i].loser] = true;
        }
    }
    return;
}

// Trying recursion to check if there's a cycle - boolean to tell if true or false
bool cycle(int winner, int loser)
{
    // Always check if our current loser is actually a winner then return true to show that there is a cycle
    // initially, if our current loser is also our current winner, then there's a loop to themselves
    if (loser == winner)
    {
        return true;
    }
    // If first case is not true then loop through candidates to find if the current loser is a locked winnner in another pair
    for (int i = 0; i < candidate_count; i++)
    {
        if (locked[loser][i] == true)
        {
            // If the candidate was indeed a locked winner then call the cycle function again with the new loser (of the pair with our previously detected locked winner)
            // winner stays the same because we are interested in the same original winner, to see if they are ever at the end of a cycle (i.e. close a cycle)
            if (cycle(winner, i))
            {
                // So, if the original winner is == to the new locked loser, return true.. else we continue with the for-loop
                return true;
            }
        }
    }
    // no true was never returned by the time we have looped through each candidate then our current winner does not close the cycle
    return false;
}

// Print the winner of the election
// Note: apparently pairs are not read in this function
void print_winner(void)
{
    // Iterate through candidates to find first winner (see line 246)
    for (int i = 0; i < candidate_count; i++)
    {
        int loser_count = 0;

        for (int j = 0; j < candidate_count; j++)
        {
            if (locked[i][j])
            {
                // Now check if the identified winner is also a loser in the locked pairs
                for (int k = 0; k < candidate_count; k++)
                {
                    // If so, add to the loser count
                    if (locked[k][i])
                    {
                        loser_count++;
                    }
                }

                // only if the loser count is still 0 by the time the loop is done, print the winning candidate!
                // Then reeturn;
                if (loser_count == 0)
                {
                    printf("%s\n", candidates[i]);
                    return;
                }
            }

        }
    }
    return;
}